/***************************************************************************/
/*                                                                         */
/* File: main.cpp                                                          */
/* URL : www.xbdev.net                                                     */
/* Email: bkenwright@xbdev.net                                             */
/*                                                                         */
/***************************************************************************/
/*                                                                         */
/* Short Description:                                                      */
/* Basic file for testing/demonstrationg collision detectoin between       */
/* a sphere and a object bounding box (OBB).                               */
/* As sometimes its easier to do the maths, but it helps if you can code   */
/* it up and actually see it in action ;)                                  */
/*                                                                         */
/***************************************************************************/

#include <windows.h>
#include <stdio.h>

// These few lines are compiler directives that include the DirectX library's
// during linking.  You could instead inside visual studio goto Project->
// settings menu and under the "Link" tabe add the necessary librarys instead.
#pragma comment(lib, "D3d8.lib")   // DirectX 8
#pragma comment(lib, "D3dx8.lib")
#include <d3dx8.h>

#include "dxdbg.h" // little mixture of mini draw functions/debug output etc


static D3DXVECTOR3 g_pos = D3DXVECTOR3(200,200,0.0f);


// DirectX Globals
LPDIRECT3D8 g_pD3D = NULL;
LPDIRECT3DDEVICE8 g_pD3DDevice = NULL;

/***************************************************************************/

// Given point p, return point q on (or in) OBB b, closest to p
void ClosestPtPointOBB(D3DXVECTOR3 p,
					   D3DXVECTOR3 bc,
					   D3DXVECTOR3* bu,
					   float* be,
					   D3DXVECTOR3 &q )
{
	D3DXVECTOR3 d = p - bc;

	// Start result at center of box; make steps from there
    q = bc;
    // For each OBB axis...
    for (int i = 0; i < 3; i++) 
	{
        // ...project d onto that axis to get the distance
        // along the axis of d from the box center
        float dist = D3DXVec3Dot(&d, &bu[i]);
        // If distance farther than the box extents, clamp to the box
        if (dist > be[i]) dist = be[i];
        if (dist < -be[i]) dist = -be[i];
        // Step that distance along the axis to get world coordinate
        q += dist * bu[i];
    }
}


// Returns true if sphere s intersects OBB b, false otherwise.
// The point p on the OBB closest to the sphere center is also returned
int TestSphereOBB_2D(	D3DXVECTOR3 sc, float sr, 
						D3DXVECTOR3 *boxCorners, 
						D3DXVECTOR3 &p)
{
	D3DXVECTOR3 bMin = D3DXVECTOR3(0,0,0);
	D3DXVECTOR3 bMax = D3DXVECTOR3(0,0,0);
	D3DXVECTOR3 obbCentre;

	int i = 0;
	bMin.x = boxCorners[i].x;
	bMin.y = boxCorners[i].y;
	bMin.z = boxCorners[i].z;
	bMax.x = boxCorners[i].x;
	bMax.y = boxCorners[i].y;
	bMax.z = boxCorners[i].z;

	for (i=0; i<4; i++)
	{
		bMin.x = min(bMin.x, boxCorners[i].x);
		bMin.y = min(bMin.y, boxCorners[i].y);
		bMin.z = min(bMin.z, boxCorners[i].z);
		bMax.x = max(bMax.x, boxCorners[i].x);
		bMax.y = max(bMax.y, boxCorners[i].y);
		bMax.z = max(bMax.z, boxCorners[i].z);
	}
	obbCentre = (bMax - bMin)/2.0f;

	float be[3];
	D3DXVECTOR3 bu[3];
	bu[0] = boxCorners[0] - boxCorners[1];
	be[0] = D3DXVec3Length(&bu[0]);
	D3DXVec3Normalize(&bu[0], &bu[0]);

    bu[1] = boxCorners[1] - boxCorners[2];
	be[1] = D3DXVec3Length(&bu[1]);
	D3DXVec3Normalize(&bu[1], &bu[1]);

	bu[2] = D3DXVECTOR3(0,0,0);
	be[2] = 0.0f;

    // Find point p on OBB closest to sphere center
    ClosestPtPointOBB(sc, 
					  obbCentre, 
					  bu,
					  be,
					  p);

    // Sphere and OBB intersect if the (squared) distance from sphere
    // center to point p is less than the (squared) sphere radius
    D3DXVECTOR3 v = p - sc;
    return D3DXVec3Dot(&v, &v) <= sr * sr;
}

/***************************************************************************/


void DrawWireBox(D3DXVECTOR3 p0,
				 D3DXVECTOR3 p1,
				 D3DXVECTOR3 p2,
				 D3DXVECTOR3 p3,
				 DWORD colour)
{
	DrawLine(g_pD3DDevice, p0.x,  p0.y,  p1.x, p1.y,  colour);
	DrawLine(g_pD3DDevice, p1.x,  p1.y,  p2.x, p2.y,  colour);
	DrawLine(g_pD3DDevice, p2.x,  p2.y,  p3.x, p3.y,  colour);
	DrawLine(g_pD3DDevice, p3.x,  p3.y,  p0.x, p0.y,  colour);

	/*
	int i = 0;
	DrawLine(g_pD3DDevice, box[i].x,  box[i].y,  box[i+1].x, box[i+1].y, colour);
	i++;
	DrawLine(g_pD3DDevice, box[i].x,  box[i].y,  box[i+1].x, box[i+1].y, colour);
	i++;
	DrawLine(g_pD3DDevice, box[i].x,  box[i].y,  box[i+1].x, box[i+1].y, colour);
	i++;
	DrawLine(g_pD3DDevice, box[i].x,  box[i].y,  box[0].x, box[0].y,     colour);
	*/
}

/***************************************************************************/
/*                                                                         */
/*  Name: KeyBoard()                                                       */
/*  Date: 03-12-05                                                         */
/*  Auth: bkenwright@xbdev.net                                             */
/*                                                                         */
/*  Used to accept keyboard input and alter some global variables - used   */
/*  for simple analysis of our code - turning wire frame rendering on/off  */
/*  adjusting variables which are used in the generation of our shape etc. */
/*                                                                         */
/***************************************************************************/
void KeyBoard()
{
	// Controlling the mesh detail
	if( GetKeyState(VK_RIGHT)	& 0x80 )  g_pos.x=min(550,g_pos.x+1);
	if( GetKeyState(VK_LEFT)	& 0x80 )  g_pos.x=max(0,  g_pos.x-1);

	if( GetKeyState(VK_UP)		& 0x80 )  g_pos.y=min(550,g_pos.y-1);
	if( GetKeyState(VK_DOWN)	& 0x80 )  g_pos.y=max(0,  g_pos.y+1);
}

// Display code
void Render()
{
	if(!g_pD3DDevice)return;


	// Clear the back buffer to a blue color
	g_pD3DDevice->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0,0,255), 1.0f, 0 );

	// Begin the scene.
	g_pD3DDevice->BeginScene();

//camera
	D3DXMATRIX view_matrix, matProj;
	D3DXMatrixLookAtLH(&view_matrix,&D3DXVECTOR3( 0.0f, 0.0f,-8.0f ),
                                   &D3DXVECTOR3( 0.0f, 0.0f, 1.0f ),
                                   &D3DXVECTOR3( 0.0f, 1.0f, 0.0f ));
	g_pD3DDevice->SetTransform(D3DTS_VIEW,&view_matrix);
	D3DXMatrixPerspectiveFovLH(&matProj, //Result Matrix
                              D3DX_PI/4,//Field of View, in radians. (PI/4) is typical
                              ((float)600 / (float)400),     //Aspect ratio
                              1.0f,     //Near view plane
                              100.0f ); // Far view plane
	g_pD3DDevice->SetTransform( D3DTS_PROJECTION, &matProj );
//end camera

	const float scale = 50.0f;
	const D3DXVECTOR3 off = D3DXVECTOR3(100,100,0);

	D3DXVECTOR3 box[4];
	box[0].x=-1.0f*scale+off.x;  box[0].y= 1.0f*scale+off.y;  box[0].z= 1.0f;
	box[1].x= 1.0f*scale+off.x;  box[1].y= 1.0f*scale+off.y;  box[1].z= 1.0f;
	box[2].x= 1.0f*scale+off.x;  box[2].y=-1.0f*scale+off.y;  box[2].z= 1.0f;
	box[3].x=-1.0f*scale+off.x;  box[3].y=-1.0f*scale+off.y;  box[3].z= 1.0f;

	DWORD hitColour = 0xff00ff00;

	D3DXVECTOR3 p;
	int hit = 0;
	hit =
	TestSphereOBB_2D(g_pos, 2, box, p);

	if (hit > 0)
	{
		hitColour = 0xffff00ff;
	}

	int i = 0;
	DrawLine(g_pD3DDevice, box[i].x,  box[i].y,  box[i+1].x, box[i+1].y, hitColour);
	i++;
	DrawLine(g_pD3DDevice, box[i].x,  box[i].y,  box[i+1].x, box[i+1].y, hitColour);
	i++;
	DrawLine(g_pD3DDevice, box[i].x,  box[i].y,  box[i+1].x, box[i+1].y, hitColour);
	i++;
	DrawLine(g_pD3DDevice, box[i].x,  box[i].y,  box[0].x, box[0].y,     hitColour);


	FlatSquare(g_pD3DDevice, g_pos.x, g_pos.y, 2, 2, hitColour);

	// End the scene.
	g_pD3DDevice->EndScene();

	// After rendering the scene we display it.
	g_pD3DDevice->Present( NULL, NULL, NULL, NULL );

}


// The mainloop() function here is just called repeately from the main.cpp file...e.g.
// its a loop so that when no messages are being processed by windows.. we come here
// you could have just put a while loop... but how would you exit the program!  Must play
// nice with the other programs on your computer as well!.
void mainloop()
{
	KeyBoard();

	Render();
}

/***************************************************************************/


void init(HWND hWnd)
{
    //First of all, create the main D3D object. If it is created successfully we 
    //should get a pointer to an IDirect3D8 interface.
    g_pD3D = Direct3DCreate8(D3D_SDK_VERSION);
	if( g_pD3D==NULL )
	{
		ErrMsg("Error Creating DirectX Object Instance");
		return;
	}

    //Get the current display mode
    D3DDISPLAYMODE d3ddm;
	HRESULT hr = 
    g_pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &d3ddm);
	if( hr!=S_OK )
	{
		ErrMsg("Failed Getting Display Mode");
		return;
	}

    //Create a structure to hold the settings for our device
    D3DPRESENT_PARAMETERS d3dpp; 
    ZeroMemory(&d3dpp, sizeof(d3dpp));
    //Fill the structure. 
    //We want our program to be windowed, and set the back buffer to a format
    //that matches our current display mode
    d3dpp.Windowed				 = TRUE;
    d3dpp.SwapEffect			 = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat		 = d3ddm.Format;
    //For depth buffering (e.g.) the z-buffer
    d3dpp.BackBufferCount		 = 1;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
    d3dpp.EnableAutoDepthStencil = TRUE;
    
	//Create a Direct3D device
	hr =														// Error Checking
    g_pD3D->CreateDevice(D3DADAPTER_DEFAULT,					// UINT Adapter
						 D3DDEVTYPE_HAL,						// D3DDEVTYPE DeviceType
						 hWnd,									// HWND hFocusWindow
						 D3DCREATE_SOFTWARE_VERTEXPROCESSING,	// DWORD BehaviorFlags
						 &d3dpp,								// D3DPRESENT_PARAMETERS *
						 &g_pD3DDevice);						// IDirect3DDevice9 *
	if( hr!=S_OK )
	{
		ErrMsg("Failed Creating Rendering Device");
		return;
	}
 
	
	//g_pD3DDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
	//g_pD3DDevice->SetTextureStageState(0,D3DTSS_COLORARG1, D3DTA_TEXTURE);
	//g_pD3DDevice->SetTextureStageState(0,D3DTSS_ALPHAOP, D3DTOP_DISABLE);
	//g_pD3DDevice->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTEXF_LINEAR);
	//g_pD3DDevice->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTEXF_LINEAR);

	//Turn on back face culling. This is becuase we want to hide the back of our polygons
    //g_pD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

	//Turn off lighting becuase we are specifying that our vertices have colour
    g_pD3DDevice->SetRenderState(D3DRS_LIGHTING, FALSE);

	//Turn on z-buffering
	g_pD3DDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);


}

void de_init()
{
	g_pD3DDevice->Release();
	g_pD3DDevice = NULL;

	g_pD3D->Release();
	g_pD3D = NULL;
}



/***************************************************************************/


// Handle all messages for the main window.
long _stdcall MsgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    if(uMsg == WM_DESTROY)
    {
		// inside init.cpp we destroy any directx memory we allocated etc, 
		// tidy up before leaving.
		de_init();
        PostQuitMessage(0);
    }
    return (long)DefWindowProc(hWnd, uMsg, wParam, lParam);
}


// Program entry point.

int _stdcall WinMain(HINSTANCE i, HINSTANCE, char* k, int) 
{
    MSG msg;
	char szname[] = "SimpleDemo";

    WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L, 
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      szname, NULL };
    RegisterClassEx( &wc );


	/*
	WNDCLASSEX WndClsEx;

		// Create the application window
	WndClsEx.cbSize        = sizeof(WNDCLASSEX);
	WndClsEx.style         = CS_HREDRAW | CS_VREDRAW;
	WndClsEx.lpfnWndProc   = MsgProc;
	WndClsEx.cbClsExtra    = 0;
	WndClsEx.cbWndExtra    = 0;
	WndClsEx.hIcon         = 0L;
	WndClsEx.hCursor       = LoadCursor(NULL, IDC_ARROW);
	WndClsEx.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	WndClsEx.lpszMenuName  = NULL;
	WndClsEx.lpszClassName = szname;
	WndClsEx.hInstance     = i;
	WndClsEx.hIconSm       = 0L;


	// Register the application
	RegisterClassEx(&WndClsEx);
*/

    HWND hWnd = CreateWindowEx(WS_EX_APPWINDOW,
					szname, szname, 
					WS_OVERLAPPEDWINDOW,//for fullscreen make into WS_POPUP
					50, 50, 500,500, //for full screen GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN),
					GetDesktopWindow(), NULL, wc.hInstance, NULL);


/*
		g_hwnd = CreateWindowEx(0,
                          szname,
                          szname,
               			  WS_SYSMENU | WS_MINIMIZEBOX,
                          CW_USEDEFAULT,
               			  CW_USEDEFAULT,
                          CW_USEDEFAULT,
               			  CW_USEDEFAULT,
                          NULL,
                          NULL,
                          WndClsEx.hInstance,
                          NULL);

*/

    //Initilise or directX code here!
    init(hWnd);

	ShowWindow(hWnd, SW_SHOW);
	UpdateWindow(hWnd);	

    // Message loop. Note that this has been modified to allow
    //  us to execute even if no messages are being processed.
    while(1)
    {
        if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
        {
            if (!GetMessage(&msg, NULL, 0, 0))
                break;
            DispatchMessage(&msg);
        }
        
        // Idle-time processing - call our loop function in game.cpp
        mainloop();
    }
    return 0;
}
